@@ -34,13 +34,13 @@ The main interface is :class:`PrestoQuery`: ::
 """
 from __future__ import absolute_import, division, print_function
 
+import logging
 import os
 from typing import Any, Dict, List, Optional, Text, Tuple, Union  # NOQA for mypy types
-import six.moves.urllib_parse as parse
 
-import prestodb.logging
 import prestodb.redirect
 import requests
+import six.moves.urllib_parse as parse
 from prestodb import constants, exceptions
 from prestodb.transaction import NO_TRANSACTION
 
@@ -48,7 +48,7 @@ from prestodb.transaction import NO_TRANSACTION
 __all__ = ["PrestoQuery", "PrestoRequest"]
 
 
-logger = prestodb.logging.get_logger(__name__)
+logger = logging.getLogger(__name__)
 
 
 MAX_ATTEMPTS = constants.DEFAULT_MAX_ATTEMPTS
@@ -62,10 +62,10 @@ else:
 class ClientSession(object):
     def __init__(
         self,
-        catalog,
-        schema,
-        source,
         user,
+        catalog = None,
+        schema = None,
+        source = None,
         properties=None,
         headers=None,
         transaction_id=None,
@@ -80,6 +80,9 @@ class ClientSession(object):
         self._headers = headers or {}
         self.transaction_id = transaction_id
 
+    def __repr__(self):
+        return f"ClientSession({self.catalog}, {self.schema}, {self.source}, {self.user}, {self._properties}, {self._headers}, {self.transaction_id})"
+
     @property
     def properties(self):
         return self._properties
@@ -199,11 +202,7 @@ class PrestoRequest(object):
         self,
         host,  # type: Text
         port,  # type: int
-        user,  # type: Text
-        source=None,  # type: Text
-        catalog=None,  # type: Text
-        schema=None,  # type: Text
-        session_properties=None,  # type: Optional[Dict[Text, Any]]
+        client_session,  # type: ClientSession
         http_session=None,  # type: Any
         http_headers=None,  # type: Optional[Dict[Text, Text]]
         transaction_id=NO_TRANSACTION,  # type: Optional[Text]
@@ -216,16 +215,7 @@ class PrestoRequest(object):
         service_account_file=None,
     ):
         # type: (...) -> None
-        self._client_session = ClientSession(
-            catalog,
-            schema,
-            source,
-            user,
-            session_properties,
-            http_headers,
-            transaction_id,
-        )
-
+        self._client_session = client_session
         self._host = host
         self._port = port
         self._next_uri = None  # type: Optional[Text]
@@ -248,6 +238,7 @@ class PrestoRequest(object):
             )
             self._http_session.headers.update(self.get_oauth_token())
 
+        self.prepared_statements = []
         self._http_session.headers.update(self.http_headers)
         self._exceptions = self.HTTP_EXCEPTIONS
         self._auth = auth
@@ -280,10 +271,12 @@ class PrestoRequest(object):
         headers[constants.HEADER_SCHEMA] = self._client_session.schema
         headers[constants.HEADER_SOURCE] = self._client_session.source
         headers[constants.HEADER_USER] = self._client_session.user
+        if len(self.prepared_statements) > 0:
+            headers[constants.HEADER_PREPARED_STATEMENT] = ",".join(self.prepared_statements)
 
         headers[constants.HEADER_SESSION] = ",".join(
             # ``name`` must not contain ``=``
-            "{}={}".format(name, parse.quote(value))
+            "{}={}".format(name, parse.quote(str(value)))
             for name, value in self._client_session.properties.items()
         )
 
@@ -412,7 +405,6 @@ class PrestoRequest(object):
 
         http_response.encoding = "utf-8"
         response = http_response.json()
-        logger.debug("HTTP {}: {}".format(http_response.status_code, response))
         if "error" in response:
             raise self._process_error(response["error"], response.get("id"))
 
@@ -428,6 +420,11 @@ class PrestoRequest(object):
             ):
                 self._client_session.properties[key] = value
 
+        if constants.HEADER_ADDED_PREPARE in http_response.headers:
+            self._http_session.headers[
+                constants.HEADER_PREPARED_STATEMENT
+            ] = http_response.headers[constants.HEADER_ADDED_PREPARE]
+
         self._next_uri = response.get("nextUri")
 
         return PrestoStatus(
@@ -482,7 +479,6 @@ class PrestoResult(object):
             rows = self._query.fetch()
             for row in rows:
                 self._rownumber += 1
-                logger.debug("row {}".format(row))
                 yield row
 
 
@@ -542,23 +538,29 @@ class PrestoQuery(object):
         response = self._request.post(self._sql)
         status = self._request.process(response)
         self.query_id = status.id
-        self._stats.update({u"queryId": self.query_id})
+        self._stats.update({"queryId": self.query_id})
         self._stats.update(status.stats)
         self._warnings = getattr(status, "warnings", [])
         if status.next_uri is None:
             self._finished = True
         self._result = PrestoResult(self, status.rows)
+        while (
+            not self._finished and not self._cancelled
+        ):
+            self._result._rows += self.fetch()
         return self._result
 
     def fetch(self):
         # type: () -> List[List[Any]]
         """Continue fetching data for the current query_id"""
+        if self._request.next_uri is None:
+            self._finished = True
+            return []
         response = self._request.get(self._request.next_uri)
         status = self._request.process(response)
         if status.columns:
             self._columns = status.columns
         self._stats.update(status.stats)
-        logger.debug(status)
         if status.next_uri is None:
             self._finished = True
         return status.rows
