@@ -17,20 +17,21 @@ https://www.python.org/dev/peps/pep-0249/ .
 Fetch methods returns rows as a list of lists on purpose to let the caller
 decide to convert then to a list of tuples.
 """
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
+from __future__ import absolute_import, division, print_function
 
-from typing import Any, List, Optional  # NOQA for mypy types
+import binascii
 import datetime
+import logging
+import uuid
+from typing import Any, List, Optional  # NOQA for mypy types
+
+import prestodb.client
+import prestodb.exceptions
+import prestodb.redirect
 
 from prestodb import constants
-import prestodb.exceptions
-import prestodb.client
-import prestodb.logging
-import prestodb.redirect
-from prestodb.transaction import Transaction, IsolationLevel, NO_TRANSACTION
-
+from prestodb.transaction import IsolationLevel, NO_TRANSACTION, Transaction
+from prestodb.transaction import NO_TRANSACTION
 
 __all__ = ["connect", "Connection", "Cursor"]
 
@@ -38,7 +39,7 @@ __all__ = ["connect", "Connection", "Cursor"]
 apilevel = "2.0"
 threadsafety = 2
 
-logger = prestodb.logging.get_logger(__name__)
+logger = logging.getLogger(__name__)
 
 
 def connect(*args, **kwargs):
@@ -75,6 +76,7 @@ class Connection(object):
         max_attempts=constants.DEFAULT_MAX_ATTEMPTS,
         request_timeout=constants.DEFAULT_REQUEST_TIMEOUT,
         isolation_level=IsolationLevel.AUTOCOMMIT,
+        **kwargs,
     ):
         self.host = host
         self.port = port
@@ -83,6 +85,15 @@ class Connection(object):
         self.catalog = catalog
         self.schema = schema
         self.session_properties = session_properties
+        self._client_session = prestodb.client.ClientSession(
+            user,
+            catalog,
+            schema,
+            source,
+            session_properties,
+            http_headers,
+            NO_TRANSACTION,
+        )
         # mypy cannot follow module import
         self._http_session = prestodb.client.PrestoRequest.http.Session()
         self.http_headers = http_headers
@@ -141,11 +152,7 @@ class Connection(object):
         return prestodb.client.PrestoRequest(
             self.host,
             self.port,
-            self.user,
-            self.source,
-            self.catalog,
-            self.schema,
-            self.session_properties,
+            self._client_session,
             self._http_session,
             self.http_headers,
             NO_TRANSACTION,
@@ -187,6 +194,9 @@ class Cursor(object):
         self._iterator = None
         self._query = None
 
+    def __iter__(self):
+        return self._iterator
+
     @property
     def connection(self):
         return self._connection
@@ -232,10 +242,126 @@ class Cursor(object):
         raise prestodb.exceptions.NotSupportedError
 
     def execute(self, operation, params=None):
-        self._query = prestodb.client.PrestoQuery(self._request, sql=operation)
-        result = self._query.execute()
-        self._iterator = iter(result)
-        return result
+        if params:
+            assert isinstance(params, (list, tuple)), (
+                "params must be a list or tuple containing the query "
+                "parameter values"
+            )
+
+            statement_name = self._generate_unique_statement_name()
+            self._prepare_statement(operation, statement_name)
+
+            try:
+                # Send execute statement and assign the return value to `results`
+                # as it will be returned by the function
+                self._query = self._execute_prepared_statement(statement_name, params)
+                self._iterator = iter(self._query.execute())
+            finally:
+                # Send deallocate statement
+                # At this point the query can be deallocated since it has already
+                # been executed
+                # TODO: Consider caching prepared statements if requested by caller
+                self._deallocate_prepared_statement(statement_name)
+        else:
+            self._query = prestodb.client.PrestoQuery(self._request, sql=operation)
+            self._iterator = iter(self._query.execute())
+        return self
+
+    def _generate_unique_statement_name(self):
+        return "st_" + uuid.uuid4().hex.replace("-", "")
+
+    def _prepare_statement(self, statement: str, name: str) -> None:
+        sql = f"PREPARE {name} FROM {statement}"
+        query = prestodb.client.PrestoQuery(self._request, sql=sql)
+        query.execute()
+
+    def _execute_prepared_statement(self, statement_name, params):
+        sql = (
+            "EXECUTE "
+            + statement_name
+            + " USING "
+            + ",".join(map(self._format_prepared_param, params))
+        )
+        return prestodb.client.PrestoQuery(self._request, sql=sql)
+
+    def _deallocate_prepared_statement(self, statement_name: str) -> None:
+        sql = "DEALLOCATE PREPARE " + statement_name
+        query = prestodb.client.PrestoQuery(self._request, sql=sql)
+        query.execute()
+
+    def _format_prepared_param(self, param):
+        """
+        Formats parameters to be passed in an
+        EXECUTE statement.
+        """
+        if param is None:
+            return "NULL"
+
+        if isinstance(param, bool):
+            return "true" if param else "false"
+
+        if isinstance(param, int):
+            # TODO represent numbers exceeding 64-bit (BIGINT) as DECIMAL
+            return "%d" % param
+
+        if isinstance(param, float):
+            if param == float("+inf"):
+                return "infinity()"
+            if param == float("-inf"):
+                return "-infinity()"
+            return "DOUBLE '%s'" % param
+
+        if isinstance(param, str):
+            return "'%s'" % param.replace("'", "''")
+
+        if isinstance(param, bytes):
+            return "X'%s'" % param.hex()
+
+        if isinstance(param, datetime.datetime) and param.tzinfo is None:
+            datetime_str = param.strftime("%Y-%m-%d %H:%M:%S.%f")
+            return "TIMESTAMP '%s'" % datetime_str
+
+        if isinstance(param, datetime.datetime) and param.tzinfo is not None:
+            datetime_str = param.strftime("%Y-%m-%d %H:%M:%S.%f")
+            # offset-based timezones
+            return "TIMESTAMP '%s %s'" % (datetime_str, param.tzinfo.tzname(param))
+
+        # We can't calculate the offset for a time without a point in time
+        if isinstance(param, datetime.time) and param.tzinfo is None:
+            time_str = param.strftime("%H:%M:%S.%f")
+            return "TIME '%s'" % time_str
+
+        if isinstance(param, datetime.time) and param.tzinfo is not None:
+            time_str = param.strftime("%H:%M:%S.%f")
+            # offset-based timezones
+            return "TIME '%s %s'" % (time_str, param.strftime("%Z")[3:])
+
+        if isinstance(param, datetime.date):
+            date_str = param.strftime("%Y-%m-%d")
+            return "DATE '%s'" % date_str
+
+        if isinstance(param, list):
+            return "ARRAY[%s]" % ",".join(map(self._format_prepared_param, param))
+
+        if isinstance(param, tuple):
+            return "ROW(%s)" % ",".join(map(self._format_prepared_param, param))
+
+        if isinstance(param, dict):
+            keys = list(param.keys())
+            values = [param[key] for key in keys]
+            return "MAP({}, {})".format(
+                self._format_prepared_param(keys), self._format_prepared_param(values)
+            )
+
+        if isinstance(param, uuid.UUID):
+            return "UUID '%s'" % param
+
+        if isinstance(param, (bytes, bytearray)):
+            return "X'%s'" % binascii.hexlify(param).decode("utf-8")
+
+        raise prestodb.exceptions.NotSupportedError(
+            "Query parameter of type '%s' is not supported." % type(param)
+        )
 
     def executemany(self, operation, seq_of_params):
         raise prestodb.exceptions.NotSupportedError
